VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "CTreeNode"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False

Option Explicit

'****************************************************
'* CTreeNode.cls                                    *
'* Standalone TreeNode class                        *
'* Programmed: Konrad Rudolph, Mad Rat Pro.         *
'* Last Change: 20.1.2003 23:27                     *
'* Version: 1.0.0                                   *
'****************************************************

'----------------------------------------------------
' INFORMATION
'----------------------------------------------------
'
' This class can be used to save data in hierarchised
' tree structures. Each tree node has all abilities
' of an own tree, thus it provides an enormous flexibility.
' Examle, if you have a Tree called Tree1, a Node called
' "Node_1" with a sub-node called Node_1.1 and you want to
' give this one a sub-node, called "Node_1.1.1".
' This is the code you would use:
'
'   Tree1.Item("Node_1").Item("Node_1.1").Add "Node_1.1.1", "Value"
'
' additionally, yould make use of the standard property
' so the above statement could become
'
'   Tree1("Node_1")("Node_1.1").Add "Node_1.1.1", "Value"
'
' or, further:
'
'   Tree1!Node_1![Node_1.1].Add "Node_1.1.1", "Value".
'
' The class supports also path finding, this means you can
' enter all keys in one string, with backslash as separator.
' The following statement would to the same as above:
'
'   Tree1.Item("Node_1\Node_1.1").Add "Node_1.1.1", "Value"
'
' notice, however, that the last way is minutely slower due
' to the parsing progress of the key string.
' The above methods (including path finding) work with all
' functions, provided by the CTreeNode class, that require
' a 'key' parametre.
'
'----------------------------------------------------

Private mValue      As String           ' the node's value
Private mKey        As String           ' a level-unique identifier

Private mCount      As Long             ' number of first-level children

Private mNext_      As CTreeNode        ' pointer to next node on same level
Private mPrev_      As CTreeNode        ' pointer to previous node on same level

Private mFirst      As CTreeNode        ' pointer to first child element
Private mLast       As CTreeNode        ' pointer to last child element

Private mParent     As CTreeNode        ' pointer to parent node
Private mChildren   As CTreeNode        ' pointer to children
'----------------------------------------------------

' returns the number of first-level child nodes
Public Property Get Count() As Long
    Count = mCount
End Property

' sets or returns the node's value
Public Property Get Value() As String
    Value = mValue
End Property
Public Property Let Value(Value As String)
    mValue = Value
End Property

'same for key
Public Property Get Key() As String
    Key = mKey
End Property
Friend Property Let Key(Value As String)
    mKey = Value
End Property

' same for parental node
Public Property Get Parent() As CTreeNode
    Set Parent = mParent
End Property
Friend Property Set Parent(Value As CTreeNode)
    Set mParent = Value
End Property

' same for next node on same tree level
Public Property Get Next_() As CTreeNode
    Set Next_ = mNext_
End Property
Friend Property Set Next_(Value As CTreeNode)
    Set mNext_ = Value
End Property

'  same for previous node on same tree level
Public Property Get Prev_() As CTreeNode
    Set Prev_ = mPrev_
End Property
Friend Property Set Prev_(Value As CTreeNode)
    Set mPrev_ = Value
End Property

' same for first node on this tree level
Public Property Get First() As CTreeNode
    Set First = mFirst
End Property
Friend Property Set First(Value As CTreeNode)
    Set mFirst = Value
End Property

' same for last node on this tree level
Public Property Get Last() As CTreeNode
    Set Last = mLast
End Property
Friend Property Set Last(Value As CTreeNode)
    Set mLast = Value
End Property

' returns the tree structure that recursively
' describes all child nodes of this one
Public Property Get Nodes() As CTreeNode
    If Not mChildren Is Nothing Then _
        Set Nodes = mChildren.First
End Property
Friend Property Set Nodes(Value As CTreeNode)
    If Not mChildren Is Nothing Then _
        Set mChildren.First = Value
End Property
'----------------------------------------------------
'----------------------------------------------------
' Add As CTreeNode
' Adds a new node to this one as a child
' - Key As String - provides an identifier for the node
' - Value As String - value of the new node
' returns a reference to the newly created node
'----------------------------------------------------
Public Function Add( _
  Key As String, _
  Optional Value As String _
) As CTreeNode
    Dim Prim    As String, _
        Append  As String
    
    Prim = Key
    If ResolveKey(Prim, Append) Then
        ' proceed sub-key call
        Set Add = Item(Prim).Add(Append, Value)
        Exit Function
    End If
    
    Set Add = New CTreeNode
    
    With Add
        .Value = Value
        .Key = Key
        Set .Parent = Me
    End With
    
    If mChildren Is Nothing Then _
        Set mChildren = New CTreeNode
    
    If mCount Then
        With mChildren
            Set .Last.Next_ = Add
            Set Add.Prev_ = .Last
            Set .Last = Add
        End With
    Else
        ' create first node
        With mChildren
            Set .First = Add
            Set .Last = Add
        End With
    End If
    
    mCount = mCount + 1
End Function

'----------------------------------------------------
' Remove As Boolean
' removes a specified node
' - Key As String - describes the node to remove
'   if more nodes with the same key are present only
'   the first one found is erased
' removes success
'----------------------------------------------------
Public Function Remove(Key As String) As Boolean
    Dim tmpNode As CTreeNode
    Dim Prim    As String, _
        Append  As String
    
    Prim = Key
    If ResolveKey(Prim, Append) Then
        ' proceed sub-key call
        Remove = Item(Prim).Remove(Append)
        Exit Function
    End If
    
    If mCount = 0 Then _
        Exit Function
    
    Set tmpNode = mChildren.First
    
    Do Until tmpNode Is Nothing
        If tmpNode.Key = Key Then
            If Not tmpNode.Next_ Is Nothing Then
                If Not tmpNode.Prev_ Is Nothing Then
                    Set tmpNode.Prev_.Next_ = tmpNode.Next_
                    Set tmpNode.Next_.Prev_ = tmpNode.Prev_
                Else
                    Set mChildren.First = tmpNode.Next_
                    Set tmpNode.Next_.Prev_ = Nothing
                End If
            Else
                If Not tmpNode.Prev_ Is Nothing Then
                    Set tmpNode.Prev_.Next_ = Nothing
                    Set mChildren.Last = tmpNode.Next_
                Else
                    Set mChildren.First = Nothing
                    Set mChildren.Last = Nothing
                End If
            End If
            
            Set tmpNode = Nothing
            Remove = True
            mCount = mCount - 1
        Else
            Set tmpNode = tmpNode.Next_
        End If
    Loop
End Function

'----------------------------------------------------
' Item As CTreeNode
' returns a reference to a specified item
' - Key As String - a specifier
' returns the first child item found with specified key
'----------------------------------------------------
' Procedure-Attribute (Default)
Public Function Item(Key As String) As CTreeNode
Attribute Item.VB_UserMemId = 0
    Dim tmpNode As CTreeNode
    Dim Prim    As String, _
        Append  As String
    
    If mChildren Is Nothing Then _
        Exit Function
    
    Prim = Key
    If ResolveKey(Prim, Append) Then
        ' proceed sub-key call
        Set Item = Item(Prim).Item(Append)
        Exit Function
    End If
    
    Set tmpNode = mChildren.First
    
    Do Until tmpNode Is Nothing
        If tmpNode.Key = Prim Then
            If LenB(Append) Then
                Set Item = tmpNode.Item(Append)
            Else
                Set Item = tmpNode
            End If
            
            Set tmpNode = Nothing
        Else
            Set tmpNode = tmpNode.Next_
        End If
    Loop
End Function

'----------------------------------------------------
' Clear
' erases whole sub-tree
'----------------------------------------------------
Public Sub Clear()
    Set mChildren = New CTreeNode
    mCount = 0
End Sub

'----------------------------------------------------
'----------------------------------------------------
' private helper function
'----------------------------------------------------

Private Function ResolveKey( _
  ByRef Key As String, _
  ByRef Append As String _
) As Boolean
    Dim Slash   As Long
    
    Slash = InStrRev(Key, "\")
    If Slash Then
        ' key contains subkeys
        
        ' remove trailing backslashes
        If RightB$(Key, 2) = "\" Then _
            Key = LeftB$(Key, LenB(Key) - 2)
        
        ' split key
        Append = MidB$(Key, Slash * 2 + 1, LenB(Key) - Slash * 2)
        Key = MidB$(Key, 1, Slash * 2 - 2)
        
        ResolveKey = True
    End If
End Function
